# TODO Thread safety
# TODO Raise exceptions on invalid input
# TODO Possible normalizations infinite loop
# TODO Comments
# TODO (attribute)_change inspection and handling
# TODO Readme file
# TODO License file

class DactylForm

  include ActiveModel::Validations

  attr_reader :object

  def initialize(object, params = nil)
    @object = object
    @params = params
    @attribute_cache = {}
    @persisted = object.persisted?
    self.class.define_getters unless self.class.getters_defined?
  end

  def http_method
    updating? ? :put : :post
  end

  def updating?
    persisted?
  end

  def creating?
    !updating?
  end

  def process
    if valid?
      save_object
      true
    else
      restore_params
      false
    end
  end

  private

  def self.object(name)
    define_method(name) { @object }
  end

  def self.attributes(*attributes)
    @attributes ||= []
    @attributes += attributes
  end

  def self.define_getters
    raise 'No attributes specified' if @attributes.blank?
    @attributes.uniq.each { |attribute| define_getter(attribute) }
    @getters_defined = true
  end

  def self.define_getter(attribute)
    define_method(attribute) do
      @attribute_cache.fetch(attribute) do

        if params_provided?
          value = read_from_params(attribute)
        elsif persisted?
          value = read_from_object(attribute)
        else
          value = read_from_default(attribute)
        end
        @attribute_cache[attribute] = value
      end
    end
  end

  def self.getters_defined?
    !!@getters_defined
  end

  def params_provided?
    !!@params
  end

  def persisted?
    @persisted
  end

  def assign_attributes
    raise "Form object should implement 'assign_attributes'"
  end

  def read_from_object(attribute)
    getter = "#{attribute}_from_object"
    respond_to?(getter) ? send(getter) : @object.send(attribute)
  end

  def read_from_default(attribute)
    getter = "#{attribute}_from_default"
    respond_to?(getter) ? send(getter) : nil
  end

  def read_from_params(attribute)
    getter = "#{attribute}_from_params"
    return @params[attribute] unless respond_to?(getter)

    arity = method(getter).arity
    arity == 0 ? send(getter) : send(getter, @params[attribute])
  end

  def save_object
    assign_attributes
    ActiveRecord::Base.transaction do
      @object.save!
      updating? ? try(:after_update) : try(:after_create)
    end
  end

  def restore_params
    @attribute_cache.each_key do |attribute|
      @attribute_cache[attribute] = @params[attribute]
    end
  end
end
